import { getPlayerProgress } from "./playerprogress";
import { getStateManager } from "./statemanager";
import { getThralls } from "./thralls";
import { SaveFileV1, mustBeSaveFileV1 } from "./saveformat";

export interface SaveFile {
  version: string;
  revision: number;
}

type SaveSlot = "FLEDGLING_SLOT_1" | "FLEDGLING_SLOT_2";

/// The result of attempting to load a V1 save file.
interface SaveFileV1LoadResult {
  // If present and valid, the loaded file.
  file: SaveFileV1 | null;

  /// A file loading error, if any. If `file` is present, this refers
  /// to an error reading from the *other* slot.
  error: string | null;

  /// The slot this file was loaded from, or that a load attempt failed from.
  /// If multiple load attempts failed and none succeeded, this refers to
  /// any one attempted slot.
  slot: SaveSlot;
}

function readFromSlot(slot: SaveSlot): SaveFileV1LoadResult {
  var serialized = localStorage.getItem(slot);
  if (serialized === null) {
    return {
      file: null,
      error: null,
      slot: slot,
    };
  }
  try {
    return {
      file: mustBeSaveFileV1(JSON.parse(serialized)),
      error: null,
      slot: slot,
    };
  } catch (e) {
    let message = "unidentifiable error";
    if (e instanceof Error) {
      message = e.message;
    }
    return {
      file: null,
      error: message,
      slot: slot,
    };
  }
}

/// Reads the more recent valid save file, if either is valid. If no save files
/// are present, `file` and `error` will both be absent. If an invalid save
/// file is discovered, `error` will refer to the issue(s) detected.
function readBestSave(): SaveFileV1LoadResult {
  const from1 = readFromSlot("FLEDGLING_SLOT_1");
  const from2 = readFromSlot("FLEDGLING_SLOT_2");
  if (from1.file && from2.file) {
    return from1.file.revision > from2.file.revision ? from1 : from2;
  }

  var errors: string[] = [];
  if (from1.error) {
    errors = ["slot 1 error: " + from1.error];
  }
  if (from2.error) {
    errors.push("slot 2 error: " + from2.error);
  }
  var msg: string | null = errors.length > 0 ? errors.join("\n") : null;
  if (from1.file) {
    return {
      file: from1.file,
      error: msg,
      slot: "FLEDGLING_SLOT_1",
    };
  }
  return {
    file: from2.file,
    error: msg,
    slot: "FLEDGLING_SLOT_2",
  };
}

export function save() {
  const targetSlot: SaveSlot =
    readBestSave().slot === "FLEDGLING_SLOT_1"
      ? "FLEDGLING_SLOT_2"
      : "FLEDGLING_SLOT_1";
  return saveIntoSlot(targetSlot);
}

function extractCurrentState(): SaveFileV1 {
  const progress = getPlayerProgress();
  const stateManager = getStateManager();
  var thrallDamage: number[] = [];
  const nThralls = getThralls().length;
  for (let i = 0; i < nThralls; ++i) {
    thrallDamage.push(progress.getThrallDamage({ id: i }));
  }
  return {
    version: "fledgling_save_v1",
    revision: stateManager.nextRevision(),
    turn: stateManager.getTurn(),
    name: progress.name,
    thrallTemplateId: progress.template.id,
    nImprovements: progress.nImprovements,
    stats: {
      agi: progress.getStat("AGI"),
      int: progress.getStat("INT"),
      cha: progress.getStat("CHA"),
      psi: progress.getStat("PSI"),
    },
    talents: {
      agi: progress.getStat("AGI"),
      int: progress.getStat("INT"),
      cha: progress.getStat("CHA"),
      psi: progress.getStat("PSI"),
    },
    isInPenance: progress.isInPenance,
    wishId: progress.getWish()?.id ?? -1,
    exp: progress.getExperience(),
    blood: progress.getBlood(),
    itemsPurloined: progress.getItemsPurloined(),
    skillsLearned: progress.getRawLearnedSkills(),
    untrimmedSkillsAvailableIds: progress.getUntrimmedAvailableSkillIds(),
    thrallsUnlocked: progress.getUnlockedThrallIds(),
    thrallDamage: thrallDamage,
    thrallsObtainedItem: progress.getThrallObtainedItemIds(),
    thrallsDeliveredItem: progress.getThrallDeliveredItemIds(),
  };
}

function saveIntoSlot(slot: SaveSlot) {
  localStorage.setItem(slot, JSON.stringify(extractCurrentState()));
}

export function wipeSaves() {
  localStorage.removeItem("FLEDGLING_SLOT_1");
  localStorage.removeItem("FLEDGLING_SLOT_2");
}
